structure Test =
struct

  exception Nope

  exception Test of string

  (* XXX urg, bad choice *)
  structure U = Util

  open SDL

  structure Util = U


  (* changing these will break things *)
  val width = 400
  val height = 400

  val max_elevation = 1000.0
  val min_elevation = 0.0
         
  val elevation: real array = Array.array (width * height, ~100.0) 

  val water_color = color(0w1,0w10,0w200,0w0)
  val shallow_water_color = color(0w40,0w40,0w220,0w0)
                    

  fun getElevation (elev, x, y) =
      Array.sub (elev, (x  + y * height) mod (width * height))

  fun setElevation (elev, x,y, el) = 
      Array.update (elev, (x + y * height) mod (width * height) , el)

  fun elevationToColor el = 
      if el < 0.0
      then if el < ~50.0 then  water_color else shallow_water_color
      else let val brightness1 = el / max_elevation
               val brightness = if brightness1 > 1.0 then 1.0 else brightness1
               val land_R = 0x32
               val land_G = 0x16
               val land_B = 0x16
               val r = Word8.fromInt( 
                       Real.floor(
                       Real.fromInt (255 - land_R) * brightness) + land_R)
               val g = Word8.fromInt(Real.floor(Real.fromInt(255 - land_G) * brightness) + land_G)
               val b = Word8.fromInt(Real.floor(Real.fromInt(255 - land_B) * brightness) + land_B)
           in color(r,g,b, 0w0)
           end


  datatype island_type = GAUSSIAN | CONE

  type island = {t: island_type, x: real, y: real, alt: real ref, width: real, dalt: real ref}

  fun square x = x * x


  val island_list: island list = 
      [  {t = GAUSSIAN, x=300.0,y=200.0, alt= ref 1000.0, width = 50.0, dalt = ref ~0.01},
         {t = GAUSSIAN, x=50.0,y=100.0, alt= ref 90.0, width = 200.0, dalt = ref 0.01},
         {t = CONE, x=110.0,y=300.0, alt= ref 200.0, width = 50.0, dalt = ref 0.0 }, 
         {t = CONE, x=350.0,y=290.0, alt= ref 80.0, width = 35.0, dalt = ref 0.02 },
         {t = CONE, x=275.0,y=350.0, alt= ref 50.0, width = 70.0, dalt = ref 0.04 },
         {t = GAUSSIAN, x=200.0, y=250.0, alt= ref 20.0, width = 100.0, dalt = ref 0.01},
         {t = GAUSSIAN, x=210.0, y=399.0, alt= ref 10.0, width = 400.0, dalt = ref 0.01}
  ]


  fun computeIsland elev x y {t, x= isle_x , y = isle_y, alt, width = isle_wide, dalt} =
      case t of
          GAUSSIAN =>
            let val dist_sq =  square (isle_x - Real.fromInt(x)) + square (isle_y - Real.fromInt(y))
                val added = !alt      * Math.exp((~ dist_sq) / square(isle_wide))
            in
                added
            end
        | CONE =>
          let val dist_sq =  square (isle_x - Real.fromInt(x)) + square (isle_y - Real.fromInt(y))
              val dist = Math.sqrt(dist_sq)
              val added = 
                  Real.max( (!alt) *( 1.0 - dist /isle_wide ), 0.0)
          in
              added
          end


 fun stepIsland ({t, x= isle_x , y = isle_y, alt, width = isle_wide, dalt}: island )= 
     alt := !(alt) + !(dalt)

 fun stepIslands isle_list = 
     let val _ = map stepIsland isle_list
     in () 
     end
 

fun sum (lst: real list): real = 
    case lst of 
        [] =>  0.0
        | x::xs =>
          (sum xs) + x


  fun computeTerrain (iter, elev, isle_list) =
    let val i  = iter mod 20
    in
        Util.for 0  (width - 1)
                 (fn x =>
                     Util.for (i*20) ((i+1) * 20 - 1)
                              (fn y =>
                                  let val els = map (computeIsland elev x y) isle_list
                                      val el1 = sum (els)
                                      val el = el1 - 100.0 (* depth of ocean *)
                                  in
                                      setElevation(elev,x,y, el)
                                  end
                          ))
    end


  type starfish = {x: int, y: int}

  val starfish_list: (starfish ref) list = 
      [ref {x = 215, y = 30}, ref { x = 390, y = 15}, ref { x = 136, y = 270}, ref {x = 20, y = 27}     ]
 


  val screen = makescreen (width, height)
  val terrain = makesurface (width, height)

  fun requireimage s =
      let val s = "testgraphics/" ^ s
      in
          case Image.load s of
              NONE => (print ("couldn't open " ^ s ^ "\n");
                       raise Nope)
            | SOME p => p
      end

  structure Font = FontFn (val surf = requireimage "font.png"
                           val charmap =
                           " ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789" ^
                           "`-=[]\\;',./~!@#$%^&*()_+{}|:\"<>?" (* \" *)
                           (* CHECKMARK ESC HEART LCMARK1 LCMARK2 BAR_0 BAR_1 BAR_2 BAR_3 
                              BAR_4 BAR_5 BAR_6 BAR_7 BAR_8 BAR_9 BAR_10 BARSTART LRARROW LLARROW *)
                           val width = 9
                           val height = 16
                           val styles = 6
                           val overlap = 1
                           val dims = 3)



  val ship_image = requireimage "ship.png"
  val starfish_image = requireimage "starfish.png"
  val shipx = ref 50.0
  val shipy = ref 150.0

  val shipdx = ref 0.0
  val shipdy = ref 0.0
                   
  val v_increment = 0.05

  val score = ref 0

  datatype intention = ACCEL_RIGHT | ACCEL_LEFT | ACCEL_UP | ACCEL_DOWN | CHILL

  fun carryOut intention = 
      case intention of
              ACCEL_RIGHT => shipdx := !shipdx + v_increment
            | ACCEL_LEFT => shipdx := !shipdx - v_increment
            | ACCEL_UP => shipdy := !shipdy + v_increment
            | ACCEL_DOWN => shipdy := !shipdy - v_increment
            | CHILL => () ;

  fun stepShip () = 
      let in 
          shipx := !shipx + !shipdx;
          shipy := !shipy + !shipdy;
          if Real.floor(! shipx) > width
          then shipx := 0.0
          else if Real.floor(!shipx) < 0
          then shipx := Real.fromInt width
          else ();

          if Real.floor(! shipy) > height
          then shipy := 0.0
          else if Real.floor(!shipy) < 0
          then shipy := Real.fromInt height
          else ()
      end

  fun stopShip () =
      let in
          shipdx := 0.0;
          shipdy := 0.0
      end
      

  fun drawShip () =
        blitall (ship_image, screen, Real.floor(!shipx), Real.floor(!shipy))

  fun drawStarfish {x,y} =
      blitall (starfish_image, screen,  x, y)
  
  fun drawStarfishes (fishes_refs: ( starfish ref) list) =
      let val fishes = map (fn x => !x) fishes_refs
      in map drawStarfish fishes
      end


  fun newStarfish (oldx: int, oldy: int) : starfish = 
    let val newx = (17 + 317 * oldx) mod width
        val newy = (13 + 129 * oldy) mod height
    in if (getElevation(elevation, newx, newy) < 0.0)
       then {x = newx, y = newy}
       else newStarfish(newx, newy)
    end


  fun checkStarfish (fish: starfish ref) = 
      let val x1 = #x (!fish)
          val y1 = #y (!fish)
          val x = Real.fromInt(x1) + 4.0
          val y = Real.fromInt(y1) + 4.0
          val myx = !shipx + 8.0
          val myy = !shipy + 15.0
          val dist = Math.sqrt ( square (x - myx) + square( y - myy))
      in if dist < 7.0
         then 
             let in 
                  score := !score + 1;
                  fish := newStarfish(x1,y1 )
              end
         else ()
      end


  fun checkStarfishes (fishes_refs: (starfish ref) list) = 
      map checkStarfish fishes_refs
      

  fun landed () =
      let val x = Real.floor(!shipx) + 8
          val y = Real.floor(!shipy) + 16
      in
          getElevation(elevation,x,y) >= 0.0
      end


  fun updateTerrain (iter, elev, isle_list) =
    let val i  = iter mod 20
    in
        Util.for 0  (width - 1)
                 (fn x =>
                     Util.for (i*20) ((i+1) * 20 - 1)
                              (fn y =>
                         drawpixel(terrain, x, y, elevationToColor(getElevation(elev, x,y)))
                          ))
    end



  fun drawElevation  () = 
      let in
          blitall (terrain, screen, 0, 0)
      end


  
 fun win() = 
      let
          val () = clearsurface (screen, water_color)
          val () = Font.draw(screen, 175, 195,  "^2YOU WIN^<")
          val () =               flip screen
      in
          case pollevent ()  of 
              SOME (E_KeyDown { sym = SDLK_SPACE }) => ()
            | _ => win()
      end



  fun dead () =
      let val () = Font.draw(screen, 140, 190,  "you hit an ^4ISLAND^<!") 
      in stopShip()
      end


  fun loop (iter, intention ) =
      let
          val next_iter = iter + 1
          val () =               clearsurface (screen, color (0w0, 0w0, 0w0, 0w0))
          val () =               computeTerrain (iter, elevation, island_list) 
          val () =               updateTerrain (iter, elevation, island_list) 
          val () =               drawElevation ()
          val () =               carryOut intention
          val _ =               drawStarfishes (starfish_list)
          val () =               stepShip ()
          val () =               drawShip ()
          val _ =               checkStarfishes (starfish_list)
          val () =               stepIslands island_list
          val () = Font.draw(screen, 10, 10,  "score: " ^ (Int.toString( !score))) 
          val () =  if landed()
                    then dead()
                    else ()
          val () =               flip screen
      in
          if !score >= 25
          then win ()
          else case pollevent ()  of 
              NONE => loop (next_iter, intention)
            | SOME (E_Quit) => ()
            | SOME (E_KeyDown { sym = SDLK_ESCAPE }) => () 
            | SOME ( E_KeyDown { sym = SDLK_RIGHT } ) =>
                  loop (next_iter, ACCEL_RIGHT)
            | SOME ( E_KeyDown { sym = SDLK_LEFT } ) =>
                  loop (next_iter, ACCEL_LEFT)
            | SOME ( E_KeyDown { sym = SDLK_UP } ) =>
                  loop (next_iter, ACCEL_DOWN)
            | SOME ( E_KeyDown { sym = SDLK_DOWN } ) =>
                  loop (next_iter, ACCEL_UP)
            | SOME ( E_KeyUp { sym } ) =>
                  loop (next_iter, CHILL)
            | _ => loop (next_iter, intention)
      end

          
 fun titleScreen() = 
      let
          val () = clearsurface (screen, water_color)
          val () = Font.draw(screen, 175, 195,  "^2GEOLOGY^<")
          val () = Font.draw(screen, 150, 230,  "press spacebar")
          val () =               flip screen
      in
          case pollevent ()  of 
              SOME (E_KeyDown { sym = SDLK_SPACE }) => ()
            | _ => titleScreen()
              
              
      end

  val () = 
      let val _ = titleScreen()
      in loop (0,CHILL)
      end

                  


end
